import json
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from functools import wraps
from typing import Any, Callable, TypeVar
from urllib.parse import ParseResult, urlparse

import click
import requests.exceptions
from mattermostdriver import Driver as Mattermost
from mattermostdriver.exceptions import MethodNotAllowed
from tabulate import tabulate

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

P = ParamSpec("P")
R = TypeVar("R")


class EmojiContext(click.Context):
    """
    Custom Click Context class
    to store global settings and manage authentication
    """

    def __init__(self) -> None:
        self.output = "table"
        self.mattermost: Any = None

    @contextmanager
    def authenticate(
        self,
        url: ParseResult,
        token: str,
        login_id: str,
        password: str,
        mfa_token: str,
        insecure: bool,
    ) -> Iterator[None]:
        """Authenticate against the Mattermost server"""

        if token and (login_id or password or mfa_token):
            click.echo(
                "Warning: Token specified along"
                " with Login-ID/Password/MFA-token."
                "Only Token will be used.",
                err=True,
            )

        settings = {
            "scheme": url.scheme,
            "url": url.hostname,
            "basepath": getattr(url, "path", ""),
            "verify": not insecure,
            "login_id": login_id,
            "password": password,
            "token": token,
            "mfa_token": mfa_token,
        }

        if url.port:
            settings["port"] = url.port
        elif url.scheme == "https":
            settings["port"] = 443
        else:
            settings["port"] = 80

        self.mattermost = Mattermost(settings)
        try:
            try:
                self.mattermost.login()
                yield
            finally:
                # Logout is unnecessary if token was used
                if not token:
                    self.mattermost.logout()
        except (requests.exceptions.ConnectionError, MethodNotAllowed) as e:
            raise click.ClickException(
                "Unable to reach Mattermost API at "
                + self.mattermost.client.url
            ) from e
        except requests.exceptions.HTTPError as e:
            raise click.ClickException(
                e.args[0] if e.args != () else repr(e)
            ) from e

    def print_dict(self, data: list[dict[str, Any]]) -> None:
        """Print dataset generated by a command to the standard output"""
        count = len(data)
        if count:
            if self.output == "table":
                click.echo(tabulate(data, headers="keys"))
            else:
                click.echo(json.dumps(data, indent=2))

        click.echo(
            f"\n({count} emoji{'' if count == 1 else 's'})",
            err=True,
        )


pass_context = click.make_pass_decorator(EmojiContext, ensure=True)


def validate_url(
    ctx: EmojiContext, param: click.Parameter, value: str
) -> ParseResult:
    """Ensure URL contains minimum information to be used"""
    url = urlparse(value)
    if not url.scheme or not url.hostname:
        raise click.BadParameter(f"Malformed URL: {value}")
    return url


def compose(
    *decorators: Callable[[Callable[P, R]], Callable[P, R]],
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Merge multiple decorators into a single one"""

    def decorate(func: Callable[P, R]) -> Callable[P, R]:
        for decorator in reversed(decorators):
            func = decorator(func)
        return func

    return decorate


global_options: Callable[[Callable[P, R]], Callable[P, R]] = compose(
    click.option(
        "-u",
        "--url",
        metavar="URL",
        envvar="MM_URL",
        callback=validate_url,
        required=True,
        help="Mattermost APIv4 URL (e.g http://localhost:8065/api/v4)"
        " (env: MM_URL)",
    ),
    click.option(
        "-t",
        "--token",
        metavar="TOKEN",
        envvar="MM_TOKEN",
        help="Personal Access Token (env: MM_TOKEN)",
    ),
    click.option(
        "-l",
        "--login-id",
        metavar="LOGIN_ID",
        envvar="MM_LOGIN_ID",
        help="login ID to use in conjunction with a password"
        " (env: MM_LOGIN_ID)",
    ),
    click.option(
        "-P",
        "--password",
        metavar="PASSWORD",
        envvar="MM_PASSWORD",
        help="password to use in conjunction with a login ID"
        " (env: MM_PASSWORD)",
    ),
    click.option(
        "-m",
        "--mfa-token",
        metavar="MFA_TOKEN",
        envvar="MM_MFA_TOKEN",
        help="optionally, Multi-Factor Authentication Token"
        " to use in conjunction with login ID/password"
        " (env: MM_MFA_TOKEN)",
    ),
    click.option(
        "-k",
        "--insecure",
        envvar="MM_INSECURE",
        is_flag=True,
        help="allow insecure server connections when using SSL"
        " (env: MM_INSECURE)",
    ),
    click.option(
        "--output",
        "-o",
        type=click.Choice(["json", "table"]),
        default="table",
        help="output format (default: table)",
    ),
)


# Workaround for the help option of subcommands not being eager enough
# The parent command is executed anyway
# https://github.com/pallets/click/issues/295
# https://github.com/pallets/click/issues/814
# This moves all global options to the subcommand which isn't that bad
# This way all options are visible when asking for help on a subcommand
def parse_global_options(
    func: Callable[P, R],
) -> Callable[P, R]:
    """Parse options used by every commands such as auth and output format"""

    @wraps(global_options(func))
    @pass_context
    def wrapper(*args: Any, **kwargs: Any) -> R:
        ctx = args[0]
        ctx.output = kwargs.pop("output")
        with ctx.authenticate(
            kwargs.pop("url"),
            kwargs.pop("token"),
            kwargs.pop("login_id"),
            kwargs.pop("password"),
            kwargs.pop("mfa_token"),
            kwargs.pop("insecure"),
        ):
            return func(*args, **kwargs)

    return wrapper



from typing import TypedDict
from urllib.parse import ParseResult, urlparse

from mmemoji.decorators import EmojiContext

from .utils import API_URL, USERS


def test_emojicontext_authenticate_password() -> None:
    observer_ctx = EmojiContext()
    ctx = EmojiContext()
    url = urlparse(API_URL)
    username = USERS["user-1"]["username"]
    password = USERS["user-1"]["password"]

    class AuthnKwargs(TypedDict):
        url: ParseResult
        token: str
        login_id: str
        password: str
        mfa_token: str
        insecure: bool

    authn_kwargs = AuthnKwargs(
        url=url,
        token="",
        login_id=username,
        password=password,
        mfa_token="",
        insecure=True,
    )

    with observer_ctx.authenticate(**authn_kwargs):
        user = observer_ctx.mattermost.users.get_user_by_username(username)
        user_id = user["id"]
        sessions_before = observer_ctx.mattermost.users.get_user_sessions(
            user_id
        )
        with ctx.authenticate(**authn_kwargs):
            sessions = ctx.mattermost.users.get_user_sessions(user_id)
            assert len(sessions) == len(sessions_before) + 1
        sessions_after = observer_ctx.mattermost.users.get_user_sessions(
            user_id
        )
        assert len(sessions_after) == len(sessions_before)



from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any, Optional, cast
from unittest.mock import _patch_dict, patch
from urllib.parse import urlparse

from mattermostdriver import Driver as Mattermost

API_URL = "http://localhost:8065/api/v4"
EMOJIS = {
    "emoji_1": {
        "path": "tests/emojis/emoji_1.png",
        "sha256": "30a8638bb79d7a99d1d8143f2679046bf7e495918fca770408011ba9579a86c7",  # noqa: E501
    },
    "emoji_2": {
        "path": "tests/emojis/emoji_2.png",
        "sha256": "725f66fba4cb6f60987f09b70d1fbd73b2f08fa4437841032103cc5300252056",  # noqa: E501
    },
    "emoji_3": {
        "path": "tests/emojis/emoji_3.png",
        "sha256": "03f4f45da7d0f8b4ee27d5961b3302b6707d1154cd2a928aeb16b107656c6a7a",  # noqa: E501
    },
    "parentheses_spaced": {
        "path": "tests/emojis/parentheses (spaced).png",
        "sha256": "195645113194074832ac56af71de520f1e2e87e52e4c8268b675832b91bab003",  # noqa: E501
    },
    "accentue": {
        "path": "tests/emojis/accentué.png",
        "sha256": "508c3f7dd47fdc0e879748cb0286e9de82d8945aab3a580c58a4a6682df6ab8f",  # noqa: E501
    },
}
USERS = {
    "sysadmin": {
        "username": "sysadmin",
        "email": "sysadmin@sample.mattermost.com",
        "password": "Sys@dmin-sample1",  # NOSONAR
    },
    "user-1": {
        "username": "user-1",
        "email": "user-1@sample.mattermost.com",
        "password": "SampleUs@r-1",  # NOSONAR
    },
}


class EmojiReconciler:
    """Maintain state of custom emojis in Mattermost"""

    def __init__(self, emoji_names: list[str], user: str) -> None:
        self.user = user
        self.emoji_names = emoji_names
        self.mattermost: Any = None
        self.authenticate()

    def authenticate(self) -> None:
        """Authenticate against the Mattermost server"""
        url = urlparse(API_URL)
        settings = {
            "scheme": url.scheme,
            "url": url.hostname,
            "port": url.port,
            "basepath": url.path,
            "login_id": USERS[self.user]["email"],
            "password": USERS[self.user]["password"],
        }

        self.mattermost = Mattermost(settings)
        self.mattermost.login()

    def create(self, name: str) -> dict[str, Any]:
        """Create emojis using a specific user"""
        with open(EMOJIS[name]["path"], "rb") as image:
            return cast(
                dict[str, Any],
                self.mattermost.emoji.create_custom_emoji(
                    name, {"image": image}
                ),
            )

    def delete(self, emoji: dict[str, Any]) -> None:
        """Delete emojis using a specific user"""
        self.mattermost.emoji.delete_custom_emoji(emoji["id"])

    def get_actual(self) -> list[dict[str, Any]]:
        """Get list of existing custom emojis on Mattermost"""
        emojis = []
        count, previous_count = 0, 0
        params = {"page": 0, "per_page": 200}
        while True:
            emojis += self.mattermost.emoji.get_emoji_list(params=params)
            count = len(emojis)
            if count - previous_count < 200:
                break
            params["page"] += 1
            previous_count = count
        return emojis

    def get_expected(self) -> list[str]:
        """Return list of desired emojis"""
        return self.emoji_names

    def reconcile(self) -> list[dict[str, Any]]:
        """Make the expected emojis match the actual emojis in Mattermost"""
        actual_emojis = self.get_actual()
        expected_emojis = self.get_expected()
        emojis = []
        for emoji in actual_emojis:
            if emoji["name"] not in expected_emojis:
                self.delete(emoji)
            else:
                emojis.append(emoji)
        emojis.extend(
            self.create(name)
            for name in expected_emojis
            if name not in [e["name"] for e in actual_emojis]
        )
        return emojis

    def destroy(self) -> None:
        """Destroy all emojis on Mattermost"""
        for emoji in self.get_actual():
            self.delete(emoji)


@contextmanager
def emoji_inventory(emoji_names: list[str], user: str) -> Iterator[None]:
    """
    Set up an inventory of emojis for the duration of a test, and then clean up
    """
    reconcilier = EmojiReconciler(emoji_names, user)
    reconcilier.reconcile()
    yield
    reconcilier.destroy()


def find_dict_in_list(
    lst: list[dict[str, Any]], key: str, value: Any
) -> Optional[dict[str, Any]]:
    """Find a dict by key name inside a list"""
    for dic in lst:
        if dic[key] == value:
            return dic
    return None


def user_env(user: str) -> _patch_dict:
    """Patch env with user credentials"""
    return patch.dict(
        "os.environ",
        {
            "MM_URL": API_URL,
            "MM_LOGIN_ID": USERS[user]["email"],
            "MM_PASSWORD": USERS[user]["password"],
        },
    )



import click
from requests import HTTPError

from mmemoji import Emoji
from mmemoji.decorators import EmojiContext, parse_global_options


@click.command(help="Search custom Emojis (200 results maximum)")
@click.argument("term")
@click.option(
    "-p",
    "--prefix-only",
    is_flag=True,
    help="only search for names starting with the search term",
)
@parse_global_options
def cli(ctx: EmojiContext, term: str, prefix_only: bool) -> None:
    try:
        ctx.print_dict(Emoji.search(ctx.mattermost, term, prefix_only))
    except HTTPError as e:
        raise click.ClickException(
            e.args[0] if e.args != () else repr(e)
        ) from e



import json
from typing import Any, cast

from click.testing import CliRunner

from mmemoji.cli import cli

from .utils import emoji_inventory, find_dict_in_list, user_env


def test_help(cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(cli, ["list", "--help"])
    assert result.exit_code == 0


def test_list_emoji(cli_runner: CliRunner) -> None:
    # Setup
    user = "user-1"
    emoji_names = ["emoji_1", "emoji_2", "emoji_3"]
    # Test
    with user_env(user), emoji_inventory(emoji_names, user):
        result = cli_runner.invoke(cli, ["list", "-o", "json"])
    assert result.exit_code == 0
    emoji_list = json.loads(result.stdout)
    emoji1 = cast(
        dict[str, Any], find_dict_in_list(emoji_list, "name", emoji_names[0])
    )
    emoji2 = cast(
        dict[str, Any], find_dict_in_list(emoji_list, "name", emoji_names[1])
    )
    emoji3 = cast(
        dict[str, Any], find_dict_in_list(emoji_list, "name", emoji_names[2])
    )
    assert len(emoji_list) == len(emoji_names)
    assert emoji1["name"] == emoji_names[0]
    assert emoji2["name"] == emoji_names[1]
    assert emoji3["name"] == emoji_names[2]



from mmemoji import Emoji


def test_emoji_sanitize_simple_name() -> None:
    name = Emoji.sanitize_name("emoji_1")
    assert name == "emoji_1"


def test_emoji_sanitize_name_with_parentheses() -> None:
    name = Emoji.sanitize_name("{[(parentheses)]}")
    assert name == "parentheses"


def test_emoji_sanitize_name_with_spaces() -> None:
    name = Emoji.sanitize_name("spa a ace")
    assert name == "spa_a_ace"


def test_emoji_sanitize_name_with_accents() -> None:
    name = Emoji.sanitize_name("àéêöhelloĐıł")  # noqa: RUF001
    assert name == "aeeohelloDil"



import json
from typing import Any, cast

from click.testing import CliRunner

from mmemoji.cli import cli

from .utils import EMOJIS, emoji_inventory, find_dict_in_list, user_env


def test_help(cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(cli, ["delete", "--help"])
    assert result.exit_code == 0


def test_delete_emoji(cli_runner: CliRunner) -> None:
    # Setup
    emoji_name = "emoji_1"
    user = "user-1"
    # Test
    with user_env(user), emoji_inventory([emoji_name], user):
        result = cli_runner.invoke(cli, ["delete", emoji_name, "-o", "json"])
    emoji_list = json.loads(result.stdout)
    emoji = cast(
        dict[str, Any], find_dict_in_list(emoji_list, "name", emoji_name)
    )
    assert result.exit_code == 0
    assert len(emoji_list) == 1
    assert emoji["name"] == emoji_name


def test_delete_absent_emoji(cli_runner: CliRunner) -> None:
    # Setup
    emoji_name = "absent_emoji"
    user = "user-1"
    # Test
    with user_env(user), emoji_inventory([], user):
        result = cli_runner.invoke(cli, ["delete", emoji_name, "-o", "json"])
    assert result.exit_code == 1
    assert result.stdout == "\n"
    error = result.stderr.split("\n")[-2]
    assert error == f'Error: Emoji "{emoji_name}" does not exist'


def test_force_delete_emoji(cli_runner: CliRunner) -> None:
    # Setup
    emoji_name = "emoji_1"
    user = "user-1"
    # Test
    with user_env(user), emoji_inventory([emoji_name], user):
        result = cli_runner.invoke(
            cli, ["delete", "--force", emoji_name, "-o", "json"]
        )
    emoji_list = json.loads(result.stdout)
    emoji = cast(
        dict[str, Any], find_dict_in_list(emoji_list, "name", emoji_name)
    )
    assert result.exit_code == 0
    assert len(emoji_list) == 1
    assert emoji["name"] == emoji_name


def test_force_delete_absent_emoji(cli_runner: CliRunner) -> None:
    # Setup
    emoji_name = "absent_emoji"
    user = "user-1"
    # Test
    with user_env(user), emoji_inventory([], user):
        result = cli_runner.invoke(
            cli, ["delete", "--force", emoji_name, "-o", "json"]
        )
    assert result.exit_code == 0
    assert result.stdout == "\n"


def test_interactive_delete_emoji(cli_runner: CliRunner) -> None:
    # Setup
    # 1st will be deleted
    # 2nd will not be deleted
    user = "user-1"
    emoji_names = ["emoji_1", "emoji_2"]
    emoji_paths = [EMOJIS[name]["path"] for name in emoji_names]
    # Test
    with user_env(user), emoji_inventory(emoji_names, user):
        result = cli_runner.invoke(
            cli,
            ["delete", "--interactive", "-o", "json", *emoji_paths],
            input="yes\nno\n",
        )
    assert result.exit_code == 0
    # Output contains the invocation input as well
    emoji_list = json.loads("\n".join(result.stdout.split("\n")[3:]))
    assert len(emoji_list) == 1
    emoji1 = cast(
        dict[str, Any], find_dict_in_list(emoji_list, "name", emoji_names[0])
    )
    assert emoji1["name"] == emoji_names[0]



from typing import TYPE_CHECKING

from mattermostdriver.exceptions import (
    InvalidOrMissingParameters,
    ResourceNotFound,
)

if TYPE_CHECKING:
    from mmemoji import Emoji


class EmojiNotFound(ResourceNotFound):  # type: ignore[misc]
    """Raised when an Emoji is not found on the Mattermost server"""

    def __init__(self, emoji: "Emoji") -> None:
        super().__init__(f'Emoji "{emoji.name}" does not exist')


class EmojiAlreadyExists(InvalidOrMissingParameters):  # type: ignore[misc]
    """Raised when an Emoji already exists on the Mattermost server"""

    def __init__(self, emoji: "Emoji") -> None:
        super().__init__(f'Emoji "{emoji.name}" exists')



import hashlib
import os
from pathlib import Path

from click.testing import CliRunner

from mmemoji.cli import cli

from .utils import EMOJIS, emoji_inventory, user_env


def test_help(cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(cli, ["create", "--help"])
    assert result.exit_code == 0


def test_download_emoji_to_directory(
    cli_runner: CliRunner, tmp_path: Path
) -> None:
    # Setup
    destination = tmp_path
    emoji_name = "emoji_1"
    emoji_filename = f"{emoji_name}.png"
    emoji_sha256 = EMOJIS[emoji_name]["sha256"]
    user = "user-1"
    # Test
    with user_env(user), emoji_inventory([emoji_name], user):
        result = cli_runner.invoke(
            cli, ["download", emoji_name, str(destination)]
        )
    paths = result.stdout.strip().split("\n")
    assert result.exit_code == 0
    assert len(paths) == 1
    assert os.path.basename(paths[0]) == emoji_filename
    with (destination / emoji_filename).open("rb") as f:
        assert hashlib.sha256(f.read()).hexdigest() == emoji_sha256


def test_download_emojis(cli_runner: CliRunner, tmp_path: Path) -> None:
    # Setup
    destination = tmp_path
    emoji_names = ["emoji_1", "emoji_2"]
    emoji_filenames = []
    emoji_sha256s = []
    for e in emoji_names:
        emoji_filenames.append(f"{e}.png")
        emoji_sha256s.append(EMOJIS[e]["sha256"])
    user = "user-1"
    # Test
    with user_env(user), emoji_inventory(emoji_names, user):
        result = cli_runner.invoke(
            cli, ["download", *emoji_names, str(destination)]
        )
    paths = result.stdout.strip().split("\n")
    assert result.exit_code == 0
    assert len(paths) == 2
    for i, path in enumerate(paths):
        assert os.path.basename(path) == emoji_filenames[i]
        with (destination / emoji_filenames[i]).open("rb") as f:
            assert hashlib.sha256(f.read()).hexdigest() == emoji_sha256s[i]


def test_download_emoji_to_full_path(
    cli_runner: CliRunner, tmp_path: Path
) -> None:
    # Setup
    destination = tmp_path / "my_emoji.img"
    emoji_name = "emoji_1"
    emoji_sha256 = EMOJIS[emoji_name]["sha256"]
    user = "user-1"
    # Test
    with user_env(user), emoji_inventory([emoji_name], user):
        result = cli_runner.invoke(
            cli, ["download", emoji_name, str(destination)]
        )
    paths = result.stdout.strip().split("\n")
    assert result.exit_code == 0
    assert len(paths) == 1
    with destination.open("rb") as f:
        assert hashlib.sha256(f.read()).hexdigest() == emoji_sha256


def test_download_emoji_to_non_existing(cli_runner: CliRunner) -> None:
    # Setup
    destination = os.path.join("path", "that", "does", "not", "exists")
    emoji_name = "emoji_1"
    user = "user-1"
    # Test
    with user_env(user), emoji_inventory([emoji_name], user):
        result = cli_runner.invoke(
            cli, ["download", emoji_name, str(destination)]
        )
    assert result.exit_code != 0
    assert not result.stdout
    assert (
        result.stderr
        == f"Error: {os.path.dirname(destination)}: Not a directory\n"
    )


def test_download_to_non_writable(
    cli_runner: CliRunner, tmp_path: Path
) -> None:
    # Setup
    destination = tmp_path
    os.chmod(destination, 0o440)
    emoji_name = "emoji_1"
    user = "user-1"
    # Test
    with user_env(user), emoji_inventory([emoji_name], user):
        result = cli_runner.invoke(
            cli, ["download", emoji_name, str(destination)]
        )
    assert result.exit_code != 0
    assert not result.stdout
    assert result.stderr == f"Error: {destination}: Permission denied\n"


def test_no_clobber_download(cli_runner: CliRunner, tmp_path: Path) -> None:
    # Setup
    destination = tmp_path
    emoji_names = ["emoji_1", "emoji_2"]
    emoji_filenames = [f"{e}.png" for e in emoji_names]
    emoji_sha256 = EMOJIS[emoji_names[1]]["sha256"]
    user = "user-1"
    # Test
    Path(destination / emoji_filenames[0]).touch()
    with user_env(user), emoji_inventory(emoji_names, user):
        result = cli_runner.invoke(
            cli,
            ["download", "--no-clobber", *emoji_names, str(destination)],
        )
    paths = result.stdout.strip().split("\n")
    assert result.exit_code == 0
    assert len(paths) == 1
    assert os.path.basename(paths[0]) == emoji_filenames[1]
    assert os.stat(destination / emoji_filenames[0]).st_size == 0
    with (destination / emoji_filenames[1]).open("rb") as f:
        assert hashlib.sha256(f.read()).hexdigest() == emoji_sha256


def test_force_download(cli_runner: CliRunner, tmp_path: Path) -> None:
    # Setup
    destination = tmp_path
    emoji_names = ["emoji_1", "emoji_2"]
    emoji_filenames = []
    emoji_sha256s = []
    for e in emoji_names:
        emoji_filenames.append(f"{e}.png")
        emoji_sha256s.append(EMOJIS[e]["sha256"])
    user = "user-1"
    # Test
    Path(destination / emoji_filenames[0]).touch()
    with user_env(user), emoji_inventory(emoji_names, user):
        result = cli_runner.invoke(
            cli, ["download", "--force", *emoji_names, str(destination)]
        )
    paths = result.stdout.strip().split("\n")
    assert result.exit_code == 0
    assert len(paths) == 2
    for i, path in enumerate(paths):
        assert os.path.basename(path) == emoji_filenames[i]
        with (destination / emoji_filenames[i]).open("rb") as f:
            assert hashlib.sha256(f.read()).hexdigest() == emoji_sha256s[i]


def test_interactive_download(cli_runner: CliRunner, tmp_path: Path) -> None:
    # Setup
    # 1st will not exist and will be downloaded
    # 2nd will exist and will be overwritten
    # 3rd will exist and will not be overwritten
    destination = tmp_path
    emoji_names = ["emoji_1", "emoji_2", "emoji_3"]
    emoji_filenames = []
    emoji_sha256s = []
    emoji_paths = []
    for e in emoji_names:
        emoji_filenames.append(f"{e}.png")
        emoji_sha256s.append(EMOJIS[e]["sha256"])
        emoji_paths.append(Path(destination / e))
    user = "user-1"
    # Test
    for e in emoji_filenames[1:]:
        Path(destination / e).touch()
    with user_env(user), emoji_inventory(emoji_names, user):
        result = cli_runner.invoke(
            cli,
            ["download", "--interactive", *emoji_names, str(destination)],
            input="yes\nno\n",
        )
    # output is sliced to exclude input from it
    paths = result.stdout.strip().split("\n")[0::2]
    assert result.exit_code == 0
    assert len(paths) == 2
    assert os.stat(destination / emoji_filenames[2]).st_size == 0
    for i, path in enumerate(paths):
        assert os.path.basename(path) == emoji_filenames[i]
        with (destination / emoji_filenames[i]).open("rb") as f:
            assert hashlib.sha256(f.read()).hexdigest() == emoji_sha256s[i]



from click.testing import CliRunner

from mmemoji import __version__
from mmemoji.cli import cli


def test_help(cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(cli, ["--help"])
    assert result.exit_code == 0


def test_version(cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(cli, ["--version"])
    assert result.exit_code == 0
    assert result.output == f"mmemoji {__version__}\n"


def test_unknown_command(cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(cli, ["unknown"])
    assert result.exit_code == 1
    assert (
        result.stderr == 'Error: Unknown command "unknown" for "mmemoji"\n'
        "Run 'mmemoji --help' for usage.\n"
    )
